@@ -1,5 +1,5 @@
 package HTTP::API::Client;
-$HTTP::API::Client::VERSION = '0.09';
+$HTTP::API::Client::VERSION = '1.0';
 use strict;
 use warnings;
 
@@ -74,90 +74,99 @@ use HTTP::Request;
 use JSON::XS;
 use LWP::UserAgent;
 use Net::Curl::Simple;
-use Mouse;
+use Moo;
 use Try::Tiny;
 use URI;
+use URI::Escape qw( uri_escape uri_unescape );
 
 has username => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
-sub _build_username { $ENV{HTTP_USERNAME} || qq{} }
+sub _build_username { _defor($ENV{HTTP_USERNAME}, '') }
 
 has password => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
-sub _build_password { $ENV{HTTP_PASSWORD} || qq{} }
+sub _build_password { _defor($ENV{HTTP_PASSWORD}, '') }
 
 has auth_token => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
-sub _build_auth_token { $ENV{HTTP_AUTH_TOKEN} || qq{} }
+sub _build_auth_token { _defor($ENV{HTTP_AUTH_TOKEN}, '') }
 
 has base_url => (
     is  => "rw",
-    isa => "URI",
 );
 
 has last_response => (
     is  => "rw",
-    isa => "HTTP::Response",
 );
 
 has charset => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
-sub _build_charset { $ENV{HTTP_CHARSET} || "utf8" }
+sub _build_charset { _defor($ENV{HTTP_CHARSET}, "utf8") }
 
 has browser_id => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_browser_id {
-    my $self = shift;
-    my $ver = $HTTP::API::Client::VERSION || -1;
+    my $ver = _defor($HTTP::API::Client::VERSION, -1);
     return "HTTP API Client v$ver";
 }
 
 has content_type => (
-    is         => "rw",
-    isa        => "Str",
-    lazy_build => 1,
+    is => "rw",
 );
 
-sub _build_content_type {
-    my $self    = shift;
+sub get_content_type {
+    my ($self, %o) = @_;
+    my $content_type = $self->content_type;
+
+    if ($content_type) {
+        return $content_type;
+    }
+
+    my $method = ${$o{method}};
+
+    if ($method eq 'GET') {
+        return 'application/x-www-form-urlencoded';
+    }
+
     my $charset = $self->charset;
     return "application/json; charset=$charset";
 }
 
 has engine => (
     is      => "ro",
-    isa     => "Str",
-    default => "LWP::UserAgent",
+    builder => 1,
 );
 
+sub _build_engine {"LWP::UserAgent"}
+
 has ua => (
-    is         => "rw",
-    isa        => "Any",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_ua {
-    my $self       = shift;
+    my ($self)     = @_;
     my $ssl_verify = $self->ssl_verify;
     my $engine     = $self->engine;
 
@@ -177,24 +186,24 @@ sub _build_ua {
 }
 
 has ssl_verify => (
-    is         => "rw",
-    isa        => "Bool",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_ssl_verify {
-    return _smart_or( $ENV{SSL_VERIFY}, 0 );
+    return _defor( $ENV{SSL_VERIFY}, 0 );
 }
 
 has retry => (
-    is         => "rw",
-    isa        => "HashRef",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_retry {
-    my $self   = shift;
-    my %retry  = %{ $self->retry_config || {} };
+    my ($self) = @_;
+    my %retry  = %{ _defor($self->retry_config, {}) };
     my $count  = $retry{fail_response};
     my %status = map { $_ => 1 } split /,/, $retry{fail_status};
 
@@ -208,45 +217,45 @@ sub _build_retry {
 }
 
 has retry_config => (
-    is         => "rw",
-    isa        => "HashRef",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_retry_config {
     return {
-        fail_response => _smart_or( $ENV{RETRY_FAIL_RESPONSE}, 0 ),
-        fail_status => $ENV{RETRY_FAIL_STATUS} || q{},
-        delay => _smart_or( $ENV{RETRY_DELAY}, 5 ),
+        fail_response => _defor( $ENV{RETRY_FAIL_RESPONSE}, 0 ),
+        fail_status => _defor($ENV{RETRY_FAIL_STATUS}, ''),
+        delay => _defor( $ENV{RETRY_DELAY}, 5 ),
     };
 }
 
 has timeout => (
-    is         => "rw",
-    isa        => "Int",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
-sub _build_timeout { return $ENV{HTTP_TIMEOUT} || 60 }
+sub _build_timeout { _defor($ENV{HTTP_TIMEOUT}, 60) }
 
 has json => (
-    is         => "rw",
-    isa        => "JSON::XS",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_json {
-    my $self    = shift;
-    my $json    = JSON::XS->new->canonical(1);
+    my ($self)  = @_;
+    my $json    = JSON::XS->new->canonical->allow_nonref;
     my $charset = $self->charset;
     eval { $json->$charset };
     return $json;
 }
 
 has debug_flags => (
-    is         => "rw",
-    isa        => "HashRef",
-    lazy_build => 1,
+    is      => "rw",
+    lazy    => 1,
+    builder => 1,
 );
 
 sub _build_debug_flags {
@@ -261,39 +270,39 @@ sub _build_debug_flags {
 
 has pre_defined_data => (
     is  => "rw",
-    isa => "HashRef",
 );
 
 has pre_defined_headers => (
     is  => "rw",
-    isa => "HashRef",
 );
 
-no Mouse;
+has pre_defined_events => (
+    is  => "rw",
+);
 
 sub get {
-    my $self = shift;
-    return $self->send( GET => @_ );
+    my ($self, @args) = @_;
+    return $self->send( GET => @args );
 }
 
 sub post {
-    my $self = shift;
-    return $self->send( POST => @_ );
+    my ($self, @args) = @_;
+    return $self->send( POST => @args );
 }
 
 sub put {
-    my $self = shift;
-    return $self->send( PUT => @_ );
+    my ($self, @args) = @_;
+    return $self->send( PUT => @args );
 }
 
 sub head {
-    my $self = shift;
-    return $self->send( HEAD => @_ );
+    my ($self, @args) = @_;
+    return $self->send( HEAD => @args );
 }
 
 sub delete {
-    my $self = shift;
-    return $self->send( DELETE => @_ );
+    my ($self, @args) = @_;
+    return $self->send( DELETE => @args );
 }
 
 sub _execute_callbacks {
@@ -309,18 +318,22 @@ sub _execute_callbacks {
 }
 
 sub send {
-    my $self         = shift;
-    my $method       = shift || "GET";
-    my $path         = shift;
-    my $data         = shift || {};
-    my $headers      = shift || {};
+    my ($self, $method, $path,
+        $data, $headers, $events) = @_;
 
-    my %options = (
-        method  => \$method,
-        path    => \$path,
-        data    => $data,
-        headers => $headers,
-    );
+    $method  = uc $method;
+    $data    = _defor( $data,    {} );
+    $headers = _defor( $headers, {} );
+    $events  = _defor( $events,  {} );
+
+    my $base_url     = $self->base_url;
+    my $url          = $base_url ? $base_url . $path : $path;
+    my $ua           = $self->ua;
+    my $retry_count  = _defor( $self->retry->{count}, 1 );
+    my $retry_delay  = _defor( $self->retry->{delay}, 5 );
+    my %retry_status = %{ _defor($self->retry->{status}, {}) };
+    my %debug        = %{ _defor($self->debug_flags, {}) };
+    my $eng          = $self->engine;
 
     if ( my $pd = $self->pre_defined_data ) {
         %$data = ( %$pd, %$data );
@@ -330,18 +343,22 @@ sub send {
         %$headers = ( %$ph, %$headers );
     }
 
+    if ( my $pe = $self->pre_defined_events ) {
+        %$events = ( %$pe, %$events );
+    }
+
+    my %options = (
+        method  => \$method,
+        url     => \$url,
+        path    => \$path,
+        data    => $data,
+        headers => $headers,
+        events  => $events,
+    );
+
     $self->_execute_callbacks(data    => %options);
     $self->_execute_callbacks(headers => %options);
 
-    my $ua           = $self->ua;
-    my $base_url     = $self->base_url;
-    my $url          = $base_url ? $base_url . $path : $path;
-    my $retry_count  = _smart_or( $self->retry->{count}, 1 );
-    my %retry_status = %{ $self->retry->{status} || {} };
-    my $retry_delay  = _smart_or( $self->retry->{delay}, 5 );
-    my %debug        = %{ $self->debug_flags || {} };
-    my $eng          = $self->engine;
-
     my $response;
 
   RETRY:
@@ -349,12 +366,14 @@ sub send {
         my $started_time = time;
 
         if ( $eng eq 'LWP::UserAgent' ) {
-            my $req   = $self->_request( $method, $url, $data, $headers );
+            my $req = $self->new_request( %options );
+
+            if ($events->{test_request_object}) {
+                return $req;
+            }
+
             $response = $ua->request($req);
         }
-        elsif ( $eng eq 'CURL' ) {
-            $response = $ua->$method
-        }
 
         if ( $debug{in_out} || $debug{send_out} ) {
             print STDERR "-- REQUEST --\n";
@@ -365,7 +384,7 @@ sub send {
             print STDERR "\n";
         }
 
-        my $debug_response = $debug{in_out} || $debug{response};
+        my $debug_response = _defor($debug{in_out}, $debug{response});
 
         $debug_response = 0
           if $debug{response_if_fail} && $response->is_success;
@@ -408,73 +427,142 @@ sub send {
 }
 
 sub json_response {
-    my $self     = shift;
-    my $response = shift;
-    try {
-        my $last_response = $self->last_response->decoded_content;
-        $response = $self->json->decode( $last_response || "{}" );
+    my ($self) = @_;
+
+    my $response = try {
+        my $content = _defor($self->last_response->decoded_content, '{}');
+        $self->json->decode($content);
     }
     catch {
         my $error = $_;
-        $response = { status => "error", error => $error };
+        { status => "error", error => $error };
     };
+
     return $response;
 }
 
-sub value_pair_response {
-    my $self = shift;
-    my @pairs = split /&/, $self->last_response->decoded_content || q{};
-    my $data =
-      { map { my ( $k, $v ) = split /=/, $_, 2; ( $k => $v ) } @pairs };
-    if ( my $error = "$@" ) {
-        $data = { status => "error", error => $error };
-    }
-    return $data;
+sub kvp_response {
+    my ($self) = @_;
+
+    my $content = $self->last_response->decoded_content
+        or return {};
+
+    my %data = map {
+        my ( $k, $v ) = map { uri_unescape($_) } split /=/, $_, 2;
+    } split /&/, $content;
+
+    return \%data;
 }
 
-sub _request {
-    my $self    = shift;
-    my $method  = uc shift;
-    my $url     = shift;
-    my $data    = shift;
-    my $headers = shift || {};
+sub new_request {
+    my ($self, %o) = @_;
 
-    my $create_req = sub {
-        my $uri = shift;
-        my $req = HTTP::Request->new( $method => $uri );
-        $req->content_type( $self->content_type )
-          if $method !~ /get/i;
-        if ( $self->username || $self->password ) {
-            _basic_authenticator( $req, $self->username, $self->password );
+    my ($method, $url) = map { $$_ } @o{qw(method url)};
+
+    my ($data, $headers, $events) = @o{qw(data headers events)};
+
+    my $content_type = $self->get_content_type(%o);
+
+    my $content = $self->convert_data(%o);
+
+    if ($content) {
+        if ($self->charset eq 'utf8') {
+            $content = _tune_utf8($content);
         }
-        elsif ( $self->auth_token ) {
-            $headers->{authorization} ||= $self->auth_token;
+    }
+
+    my $request;
+
+    if ($method eq 'GET') {
+        if ($content_type ne 'application/x-www-form-urlencoded') {
+            die "Unable to create a get request with content_type: $content_type";
         }
-        return $req;
-    };
-
-    my $req = $create_req->($url);
-
-    my $content = _tune_utf8( $self->_convert_data( $req, $data ) );
-
-    if ( $method =~ /get/i ) {
-        $req = $create_req->( $content ? "$url?$content" : $url );
+        elsif ($content) {
+            if ($url =~ m/\?/) {
+                $request = $self->prepare_request(%o, url => \"$url&$content");
+            }
+            else {
+                $request = $self->prepare_request(%o, url => \"$url?$content");
+            }
+        }
+        else {
+            $request = $self->prepare_request(%o);
+        }
+    }
+    elsif ($content) {
+        $request = $self->prepare_request(%o);
+        $request->content($content);
     }
 
-    foreach my $field ( keys %$headers ) {
-        $req->header( $field => $headers->{$field} );
+    %o = (%o,
+        request => $request,
+        content => \$content,
+    );
+
+    if (my $do = $events->{before_headers}) {
+        $self->$do(%o);
     }
 
-    if ( $method !~ /get/i ) {
-        $req->content($content);
+    my @keys;
+
+    if (my $keys = $events->{headers_keys}) {
+        @keys = $self->$keys(%o);
+    }
+    elsif (my $add = $events->{add_headers_keys}) {
+        @keys = sort $self->$add(%o), keys %$headers;
+    }
+    else {
+        @keys = sort keys %$headers;
     }
 
-    return $req;
+    foreach my $key ( @keys ) {
+        if (my $do = $events->{before_header}{$key}) {
+            $headers->{$key} = $self->$do(%o);
+        }
+
+        $request->header( $key => $headers->{$key} );
+
+        if (my $do = $events->{after_header}{$key}) {
+            $self->$do(%o);
+        }
+    }
+
+    if (my $do = $events->{after_header_keys}) {
+        $self->$do(%o);
+    }
+
+    return $request;
+}
+
+sub prepare_request {
+    my ($self, %o) = @_;
+
+    my ($method, $url) = map { $$_ } @o{qw(method url)};
+
+    my ($headers) = @o{qw(headers)};
+
+    my $request = HTTP::Request->new( $method => $url );
+
+    $request->content_type($self->get_content_type(%o));
+
+    my ($u, $p, $at) = map { _defor($self->$_, '') }
+        qw(username password auth_token);
+
+    if ($u || $p) {
+        $self->basic_authenticator($request, $u, $p);
+    }
+    elsif ($at) {
+        $headers->{authorization} = $at;
+    }
+
+    return $request;
 }
 
 sub _tune_utf8 {
-    my $content = shift;
+    my ($content) = @_;
+
     my $req = HTTP::Request->new( POST => "http://find-encoding.com" );
+
     try {
         $req->content($content);
     }
@@ -487,43 +575,74 @@ sub _tune_utf8 {
     return $content;
 }
 
-sub _convert_data {
-    my $self = shift;
-    my $req  = shift;
-    my $data = shift;
+sub convert_data {
+    my ($self, %o) = @_;
 
-    return $data
-      if !ref $data;
+    my ($data, $events) = @o{qw(data events)};
 
-    my $ct = $req->content_type
-      or return _hash_to_query_string(%$data);
+    my $content_type = $self->get_content_type(%o);
 
-    return $ct =~ /json/
-      ? $self->json->encode($data)
-      : _hash_to_query_string(%$data);
+    if ($content_type =~ m/json/) {
+        return $self->json->encode($data);
+    }
+    elsif ($content_type eq 'application/x-www-form-urlencoded') {
+        return $self->kvp2str(%o);
+    }
+    else {
+        return $data;
+    }
 }
 
-sub _hash_to_query_string {
-    my %hash = @_;
-    my $uri  = URI->new("http://parser.com");
-    $uri->query_form( \%hash );
-    my ( undef, $params ) = split /\?/, $uri->as_string;
-    return $params;
+sub kvp2str {
+    my ($self, %o) = @_;
+
+    my ($data, $events) = @o{qw(data events)};
+
+    my @keys;
+
+    if (my $do = $events->{before_sorting_keys}) {
+        $self->$do($data, \@keys);
+    }
+
+    if (my $do = $events->{keys}) {
+        @keys = $self->$do($data);
+    }
+    else {
+        @keys = sort keys %$data;
+    }
+
+    if (my $do = $events->{after_sorting_keys}) {
+        $self->$do($data, \@keys);
+    }
+
+    my @parts;
+
+    foreach my $key(@keys) {
+        if ($o{skip_key}{$key}) {
+            next;
+        }
+
+        my $val = _defor($data->{$key}, '');
+
+        $val = uri_escape($val);
+        $key = uri_escape($key);
+
+        push @parts, "$key=$val";
+    }
+
+    return join '&', @parts;
 }
 
-sub _basic_authenticator {
-    my $req      = shift;
-    my $username = shift;
-    my $password = shift;
-    $req->headers->authorization_basic( $username, $password );
+sub basic_authenticator {
+    my ($self, $req, $u, $p) = @_;
+    return $req->headers->authorization_basic($u, $p);
 }
 
-sub _smart_or {
-    my $default_value = shift;
-    my $or_value      = shift;
-    return
-      defined($default_value)
-      && length($default_value) ? $default_value : $or_value;
+sub _defor {
+    my ($default, $or) = @_;
+    return (defined($default) && length($default)) ? $default : $or;
 }
 
+no Moo;
+
 1;
